/* Copyright (C) 2006 Pierre Nicolas
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifndef _COALNODE_H_
#define _COALNODE_H_

#include <vector>
#include <iostream>
#include <gsl/gsl_rng.h>

using namespace std;

class Coalnode
{

 protected:
  bool _contextdep;
  int _lseq;
  int _nmodal;
  int* _seq;
  int* _seqbis;
  int* _obs_seq; // == NULL if _node_status != 0 else it points to _cells_obs_seqs[_iseq]

  int **_countseq;
  int *_countmut;
  int _ncounts;

  int _iseq; // meaningfull only for type 1 node, refers to index in _seq

  double _time;

  int _nb_mutclass;

  Coalnode* _child_right;
  Coalnode* _child_left;
  Coalnode* _ancestor;

  Coalnode* _next;
  Coalnode* _prev;


  int _node_status; // 0: leaf ; 1: internal ; 2: root

  gsl_rng * _rng;

  Mutmodel **_mut_model;

  double** _probaswitch;
  double* _posterior_probas_mutclass;

  double*** _probatrans;
  double*** _posterior_probas;
  double** _peeling_seq;
  double* _peeling_mutclass;

  int _mut_class;

  static int _count;
  static int _id;
  int _node_id;

 public:
  Coalnode(int node_status, Coalnode* prev, Coalnode* next, 
	   Coalnode* child_right, Coalnode* child_left,
	   Coalnode* ancestor, double time, int lseq, Mutmodel **mut_model, 
	   gsl_rng* rng);
  ~Coalnode();

  void setiseq(int iseq, int* obs_seq);
  int iseq();

  void drawSeq();
  void printInfo();

  void computeProbaTrans();
  
  double likelihoodNode();
  double sumLikelihoodNode();
  void posteriorSimulNode();

  double loglikSeqChildren();
  double loglikAncestor();

  void peelingSeqBackward();
  void simulSeqForward();
  void peelingMutClassBackward();
  void simulMutClassForward();
  
  void moveafter(Coalnode* previous_node);
  void movebefore(Coalnode* next_node);
  void increaseNbTermStemCell();
  void decreaseNbTermStemCell(double new_coaltime);

  void increaseNbInitStemCell(Coalnode* &node_collector, double coaltime);
  void decreaseNbInitStemCell(Coalnode* &node_collector);

  void setseq(int* seq);

  void splitTreeWB();
  void joinTreeWB(Coalnode* target, double coaltime);

  void resetCountSeq();
  void countSeq();
  void printCountSeq();
 

  inline void changeseq(int ipos);
  inline int seq(int pos);
  inline vector<int> seq();
  inline int* seq_ptr();
  inline double time();
  inline void change_time(double time);
  inline void swap_order();
  inline int node_status();

  Coalnode* diffLineageRoot();
  inline int node_id();
  inline Coalnode* next();
  inline Coalnode* prev();
  inline Coalnode* ancestor();
  inline Coalnode* child_left();
  inline Coalnode* child_right();
  inline void swap_childs();
  inline void change_child_left(Coalnode* new_child_left);
  inline void change_child_right(Coalnode* new_child_right);
  inline void change_child_next(Coalnode* new_next);
  inline void switchSeq();
  inline double distseq(Coalnode* node);
  inline int mut_class();
  inline void checkContextDep();
  inline void getChildCoaltimes(vector<double> & coaltimes);
};

inline void Coalnode::getChildCoaltimes(vector<double> & coaltimes)
{
  if (_node_status==1) 
    {
      coaltimes.push_back(_time);
    }
  if (_child_left != NULL) 
    {
      _child_left->getChildCoaltimes(coaltimes);      
    }
  if (_child_right != NULL) 
    {
      _child_right->getChildCoaltimes(coaltimes);      
    }  
}

inline void Coalnode::checkContextDep()
{
  _contextdep = (*_mut_model)->contextdep();
  if (_next != NULL)
    {
      _next->checkContextDep();
    }
}

inline int Coalnode::mut_class()
{
  return _mut_class;
}

inline double Coalnode::distseq(Coalnode* node)
{
  double res=0;
  for (int ipos=0; ipos<_lseq; ipos++)
    {
      if (_seq[ipos] != node->_seq[ipos])
	{
	  res++;
	}
    }
  return res;
}

inline void Coalnode::changeseq(int ipos)
{
  if (_seq[ipos]==0)
    {
      _seq[ipos]=1;
    } else {
      _seq[ipos]=0;
    }
}

inline void Coalnode::switchSeq()
{
  int* seqtmp = _seqbis;
  _seqbis = _seq;
  _seq = seqtmp;
  if (_next != NULL)
    {
      _next->switchSeq();
    }
}

inline void Coalnode::change_child_left(Coalnode* new_child_left)
{
  _child_left = new_child_left;
  _child_left->_ancestor = this;
}

inline void Coalnode::change_child_right(Coalnode* new_child_right)
{
  _child_right = new_child_right;
  _child_right->_ancestor = this;
  
}

inline void Coalnode::swap_order()
{
  //cout << "Swap order" << endl;
  Coalnode* node2 = _next;

  this->_prev->_next = node2;
  node2->_next->_prev = this;

  this->_next = node2->_next;
  Coalnode* tmp = this->_prev;
  this->_prev = node2;

  node2->_prev = tmp;
  node2->_next = this;
}

inline void Coalnode::swap_childs()
{
  //cout << "Swap child node " << _node_id << endl;
  Coalnode* tmp = _child_right;
  _child_right = _child_left;
  _child_left = tmp;    
}


inline void Coalnode::change_time(double time)
{
  _time = time;
}

inline int Coalnode::seq(int pos)
{
  return _seq[pos];
}

inline vector<int> Coalnode::seq()
{
  //cout << "node_id " << _node_id << endl;
  vector<int> seq(_lseq);
  for (int i=0; i<_lseq; i++) 
    {
      seq[i] = _seq[i];
    }
  return seq;
}

inline int* Coalnode::seq_ptr()
{
  return _seq;
}


inline int Coalnode::node_status()
{
  return _node_status;
}

inline int Coalnode::node_id()
{
  return _node_id;
}

inline double Coalnode::time()
{
  return _time;
}

inline Coalnode* Coalnode::next()
{
  if (_next != NULL)
    {
      //cout << "next() node_id " << _node_id << " -> " << _next->_node_id << endl;
    } else {
      //cout << "next() node_id " << _node_id << " -> " << "NULL " << endl;
    }
  return _next;
}

inline Coalnode* Coalnode::prev()
{
  return _prev;
}

inline Coalnode* Coalnode::ancestor()
{
  return _ancestor;
}

inline Coalnode* Coalnode::child_left()
{
  return _child_left;
}

inline Coalnode* Coalnode::child_right()
{
  return _child_right;
}

#endif
