/* Copyright (C) 2006 Pierre Nicolas
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifndef _CRYPT_H_
#define _CRYPT_H_

#include <fstream>

#include <gsl/gsl_rng.h>
#include <map>

#include "coalnode.h"
#include "stat_withincrypt.h"

using namespace std;

class Crypt
{

 protected:
  int _lseq;
  bool* _missing;

  int _nmodal;

  int _ncells;
  int _nstemcell_sampled;
  int** _cells_obs_seqs;
  //int** _cells_ori_seqs; // same as _cells_obs_seqs but without the sequencing errors

  int** _errors_counts;

  double* _ptarget_wb;
  Coalnode** _target_wb;

  double* _prior_proba_N;
  double* _posterior_proba_N;

  double** _stirling2;

  int _N_stem_cells;
  int _nb_internal_nodes;

  double _age_crypt;
  double _lambda;
  double _g; // differentiation lineages star-likeness parameter

  Mutmodel *_mut_model;

  Coalnode* _firstnode;
  Coalnode* _lastnode;

  Coalnode* _node_collector;

  gsl_rng * _rng;

  static int _count;
  int _crypt_id;

 public:
  Crypt(Mutmodel *mut_model, int N_stem_cells, double lambda, double g, gsl_rng * rng, double** stirling2, ifstream& input); 
  ~Crypt();
  
  inline void changeMutmodel(Mutmodel* newmodel);

  void countseq(map<vector<int>, int>& seqcounts, int** sequences);
  void drawDiffCoalTimesFromPrior(vector<double>& coaltimes, int n);
  double loglikDiffCoal(vector<double>& coaltimes);
  inline double diffRatefunc(double s);
  inline double diffIntegratedRatefunc(double s);
  double invIntegratedRateFunc(double s);

  void simulPrior();
  void drawGenealogyPrior();
  void drawSeqGenealogyPrior(); 
  void drawSeqErrorPrior(); 

  void updateSeq();
  void updateSeqForwBackw();
  void updateLevel();
  void updateSingleLevel(int level, Coalnode* _node);
  
  void updateTreeWB(bool updateN, int* nb_stemcells_counts,int& Nmin, int& Nmax, 
		    int nstemcells_prior_min, int nstemcells_prior_max);

  double computeLogLikSeq();
  double computeLogLikTree();
  double computeLogLikTree(Coalnode* diffroot1, Coalnode* diffroot2);
  double computeLogLikTreeN();
  double computeLogLikTreeN1();
  double computeLogLikN();
  void updateTermStemCell(bool fake);
  void updateInitStemCell();
  void updateNstemcell(double* nb_stemcells_distrib, int nstemcells_min, int nstemcells_max);
  void updateNstemcellFancy(int* nb_stemcells_counts, 
			    int& Nmin, int& Nmax, 
			    int nstemcells_prior_min, int nstemcells_prior_max);
  void computePriorProbaN(int* nb_stemcells_counts, 
			  int Nmin, int Nmax, int& Npriormin, int& Npriormax,
			  int nstemcells_prior_min, int nstemcells_prior_max);
  void printInfo();
  inline double lambda();
  inline void setlambda(double newlambda);
  inline void setg(double newg);
  inline int Nstemcell();
  inline bool setNstemcell(int newN);
  void printCrypt();
  void printCryptFromObsSeq();
  void printCryptFromObsSeq(ostream& out);
  double coalratefunc(int nbliving);
  inline void changeseq(int& letter);
  void resetErrorCounts();
  void storeErrorCounts();
  void printErrorCounts();
  void countErrors(int& sum1, int& sum2);
  void deleteGenealogy();
  StatWithinCrypt* summaryStat(bool simulseq);
  void printTree4R(ostream& out, string tagtree);
  inline Mutmodel* mutmodel();
  inline Coalnode* firstnode();
  inline int nstemcellsampled();
  inline int ntermstemcell();
  void checkSeqLikComputations();
  inline void changeRng(gsl_rng * rng);
  bool checkNodeOrder();
  void printCoaltimes();
  inline int ncells();
};

inline int Crypt::ncells()
{
  return _ncells;
}

inline void Crypt::changeRng(gsl_rng * rng)
{
  _rng = rng;
}

inline int Crypt::nstemcellsampled()
{
  return _nstemcell_sampled;
}

inline int Crypt::ntermstemcell()
{
  int n = 0;
  Coalnode* currentnode = _lastnode;
  while (currentnode->time()>40-0.00001) 
    //while (currentnode->node_status()==2) 
    {
      currentnode = currentnode->prev();
      n++;
    }
  return n;
}

inline Coalnode* Crypt::firstnode()
{
  return _firstnode;
}

inline void Crypt::changeMutmodel(Mutmodel* newmodel)
{
  _mut_model = newmodel;
  
}

inline Mutmodel* Crypt::mutmodel()
{
  return _mut_model;
}

inline void Crypt::changeseq(int& letter)
{
  if (letter==0)
    {
      letter=1;
    } else {
      letter=0;      
    }
}

inline double Crypt::lambda()
{
  return _lambda;
}

inline void Crypt::setlambda(double newlambda)
{
  _lambda = newlambda;
}

inline void Crypt::setg(double newg)
{
  _g = newg;
}

inline int Crypt::Nstemcell()
{
  return _N_stem_cells;
}

inline bool Crypt::setNstemcell(int newN)
{
  _N_stem_cells = newN;
  if (_nstemcell_sampled > newN)
    {
      return false;
    }
  return true;
}

inline double Crypt::diffRatefunc(double s)
{
  //return _g*log(2)*pow(2.,_g*s)/(pow(2.,_g)-pow(2._g*t));
  return _g*log(2.)/(pow(2.,_g*(1-s))-1);
}

inline double Crypt::diffIntegratedRatefunc(double s)
{
  return -log( (pow(2.,_g)-pow(2.,_g*s))/(pow(2.,_g)-1) );
}

#endif
